import importlib.util
import json
import sys
from pathlib import Path
from typing import Optional

import typer
from rich import print as rprint

from trubrics.exceptions import MissingConfigPathError, MissingTrubricRunFileError
from trubrics.utils.trubrics_manager_connector import make_request
from trubrics.validations.run import TrubricRun, run_trubric

app = typer.Typer()


@app.command()
def run(
    save_ui: bool = False,
    trubric_config_path: str = typer.Option(
        ...,
        prompt=(
            "Enter the path to your trubric config .json file (this file can be generated by running `trubrics init`)"
        ),
    ),
    trubric_output_file_path: str = typer.Option(
        ".", prompt="Enter a path to save your output trubric file. The default path is"
    ),
    trubric_output_file_name: str = typer.Option(
        "my_new_trubric.json", prompt="Enter the file name of your output trubric file. The default file name is"
    ),
):
    """The CLI `trubrics run` command for running trubrics.

    Args:
        trubric_config_path: path to your trubric config .json file \
            (this file can be generated by running `trubrics init`).
        trubric_output_file_path: path to save your output trubric file
        trubric_output_file_name: file name of your output trubric file.
    """
    trubric_config_file_path = Path(trubric_config_path) / ".trubrics_config.json"
    if not trubric_config_file_path.exists():
        raise MissingConfigPathError(
            f"The trubric configuration file '{trubric_config_file_path}' has not been found. Run `trubrics init` to"
            " create file."
        )
    with open(trubric_config_file_path, "r") as file:
        trubrics_config = json.load(file)
    trubric_run_path = trubrics_config["trubric_run_path"]

    tc = _import_module(module_path=trubric_run_path)
    if hasattr(tc, "RUN_CONTEXT"):
        if isinstance(tc.RUN_CONTEXT, TrubricRun):
            run_context = tc.RUN_CONTEXT
        else:
            raise TypeError("'RUN_CONTEXT' attribute must be of type TrubricRun.")
    else:
        raise AttributeError("Trubrics config python module must contain an attribute 'RUN_CONTEXT'.")

    typer.echo(
        typer.style(
            f"Running trubric from file '{trubric_run_path}' with model '{run_context.trubric.model_name}' and"
            f" dataset '{tc.data_context.name}'.",
            fg=typer.colors.BLUE,
        )
    )
    all_validation_results = run_trubric(tr=run_context)
    validations = []
    for validation_result in all_validation_results:
        validations.append(validation_result)

        message_start = f"{validation_result.validation_type} - {validation_result.severity.upper()}"
        completed_dots = (100 - len(message_start)) * "."
        if validation_result.outcome == "pass":
            ending = typer.style("PASSED", fg=typer.colors.GREEN, bold=True)
        else:
            ending = typer.style("FAILED", fg=typer.colors.WHITE, bg=typer.colors.RED)
        message = typer.style(message_start, bold=True) + completed_dots + ending
        typer.echo(message)

    # save new trubric .json
    new_trubric = tc.trubric
    new_trubric.validations = validations
    new_trubric.save_local(path=trubric_output_file_path, file_name=trubric_output_file_name)

    # save new trubric to ui
    if save_ui is True:
        if "user_id" in trubrics_config.keys():
            new_trubric.save_ui(url=trubrics_config["api_url"], user_id=trubrics_config["user_id"])
        else:
            typer.echo(
                typer.style(
                    "ERROR: You must authenticate with the trubrics manager by running `trubrics init` to remotely save"
                    " trubrics runs.",
                    fg=typer.colors.RED,
                )
            )


@app.command()
def init(
    trubrics_api_url: Optional[str] = None,
    trubric_run_path: str = typer.Option(
        ..., prompt="Enter the path to your trubric run .py file (e.g. examples/classification_titanic/trubric_run.py)"
    ),
    trubric_config_path: str = typer.Option(
        ".", prompt="Enter a path to save your .trubrics_config.json. The default path is"
    ),
):
    """The CLI `trubrics init` command for initialising trubrics config.

    Args:
        trubric_run_path: a path towards the .py file that holds code to run a trubric in \
            the `RUN_CONTEXT=TrubricRun(...)`.
        trubric_config_path: a path to save the .trubrics_config.json to
    """

    if trubrics_api_url:
        uid = typer.prompt("Enter your User ID (generated in the trubrics manager)")

        res = make_request(f"{trubrics_api_url}/api/is_user/{uid}", headers={"Content-Type": "application/json"})
        res = json.loads(res)
        if "is_user" in res.keys():
            message = typer.style(res["msg"], fg=typer.colors.RED, bold=True)
            typer.echo(message)
            raise typer.Abort()
        typer.echo(
            typer.style(
                "Trubrics configuration has been set and user is authenticated with the trubrics manager UI:",
                fg=typer.colors.GREEN,
                bold=True,
            )
        )
        res["api_url"] = trubrics_api_url
    else:
        typer.echo(
            typer.style(
                "Trubrics config set without trubrics manager authentication:", fg=typer.colors.GREEN, bold=True
            )
        )
        res = {}

    if not Path(trubric_run_path).exists() or not trubric_run_path.endswith(".py"):
        raise MissingTrubricRunFileError("Trubric run file path does not exist or is not a .py file.")

    res["trubric_run_path"] = trubric_run_path

    rprint(res)
    with open(
        Path(trubric_config_path) / ".trubrics_config.json",
        "w",
    ) as file:
        file.write(json.dumps(res, indent=4))


@app.command()
def example_titanic_app():
    """Run the titanic user feedback collector app."""
    import os

    import streamlit.cli

    dirname = os.path.dirname(__file__)
    filename = os.path.join(dirname, "../example/feedback_app_titanic.py")
    args = []
    streamlit.cli._main_run(filename, args)


def _import_module(module_path: str):
    try:
        spec = importlib.util.spec_from_file_location("module.name", module_path)
        lib = importlib.util.module_from_spec(spec)  # type: ignore
        sys.modules["module.name"] = lib
        spec.loader.exec_module(lib)  # type: ignore
    except FileNotFoundError as e:
        raise e
    return lib


if __name__ == "__main__":
    app()



import json

import requests  # type: ignore

from trubrics.feedback.dataclass import Feedback


def save_validation_to_json(trubric_context: Feedback, tracking: bool = False) -> None:
    test_json = trubric_context.json()
    if tracking:
        url = "http://localhost:8000"
        headers = {"Content-type": "application/json"}
        requests.post(
            url + "/api/feedback/",
            data=test_json,
            headers=headers,
        )
    else:
        with open(
            "examples/data/feedback.json",
            "w",
        ) as file:
            file.write(test_json)


def get_business_feedback_data(
    tracking: bool = False,
) -> Feedback:
    if tracking:
        raise ValueError("to be replaced with read from test tracking API")
    else:
        with open(
            "../data/feedback_demo.json",
            "r",
        ) as file:
            saved_test = json.load(file)
            return saved_test



from typing import Any, Optional, Tuple, Union

import pandas as pd
import streamlit as st
from pandas.api.types import is_numeric_dtype

from trubrics.context import DataContext, TrubricsModel
from trubrics.exceptions import PandasSchemaError
from trubrics.feedback.dataclass import Feedback
from trubrics.utils.pandas import schema_is_equal


class FeedbackCollector:
    def __init__(self, data: DataContext, model: Any):
        self.trubrics_model = TrubricsModel(data=data, model=model)
        self.model_type = self.trubrics_model.model_type
        self.what_if_df = None
        self.what_if_prediction = None

    def generate_what_if(self, wi_data: Optional[pd.DataFrame] = None):
        """
        Generate a what-if tool based on a DataFrame input.
        If no DataFrame specified, the DataContext testing data is used.

        TODO: check if total columns > N, option to select top n features based on dict of feature importance
        """
        if wi_data is None:
            wi_data = self.trubrics_model.data.X_test
        else:
            if not schema_is_equal(wi_data, self.trubrics_model.data.testing_data):
                raise PandasSchemaError("Schemas of provided data and X_test in DataContext are different.")

        out_df = pd.DataFrame()
        for col, dtype in wi_data.dtypes.to_dict().items():
            series = wi_data[col]
            if (
                self.trubrics_model.data.business_columns is not None
                and col in self.trubrics_model.data.business_columns.keys()
            ):
                renamed_col = self.trubrics_model.data.business_columns[col]
            else:
                renamed_col = col
            if self.trubrics_model.data.categorical_columns is None:
                raise ValueError(
                    "A list of categorical_columns must be specified in the DataContext for the What-If tool."
                )
            if col in self.trubrics_model.data.categorical_columns:
                if is_numeric_dtype(dtype.type):
                    out_df[col] = [
                        st.slider(
                            renamed_col,
                            min_value=int(series.min()),
                            max_value=int(series.max()),
                            value=round(series.mean()),
                        )
                    ]
                else:
                    out_df[col] = [st.selectbox(renamed_col, tuple(series.dropna().unique()))]
            elif is_numeric_dtype(dtype.type):
                out_df[col] = [
                    st.number_input(
                        renamed_col,
                        min_value=int(series.min()),
                        max_value=int(series.max()),
                        value=int(series.mean()),
                        step=None,
                    )
                ]
            else:
                raise NotImplementedError(f"Columns '{col}' type is not recognised.")
        self.what_if_df = out_df
        self.what_if_prediction = self.trubrics_model.model.predict(out_df)[0]

    def save_feedback(self, path: str, file_name: str):
        """Get user feedback and save"""
        feedback_type: str = st.selectbox(
            "Choose feedback type:",
            (
                "Single prediction error",
                "Bias",
                "Important features",
                "Other",
            ),
        )

        if self.what_if_df is None:
            st.text("What-if tool must be set to generate feedback.")
        else:
            metadata = {}
            what_if_input = self.what_if_df.to_dict("records")[0]
            if feedback_type == "Other":
                metadata["description"] = st.text_input(label="", value="Send free text feedback here")
                metadata["what_if_input"] = what_if_input
                metadata["model_prediction"] = self.what_if_prediction

            elif feedback_type == "Single prediction error":
                metadata["corrected_prediction"], metadata["description"] = self._collect_single_edge_case()
                metadata["what_if_input"] = what_if_input

            elif feedback_type == "Important features":
                (
                    metadata["selected_feature"],
                    metadata["top_n_feature"],
                    metadata["description"],
                ) = self._collect_important_features_feedback()

            elif feedback_type == "Bias":
                metadata["description"] = "Feedback on bias."

            else:
                raise NotImplementedError()

            single_test = Feedback(feedback_type=feedback_type, metadata=metadata)
            if st.button("Send feedback"):
                single_test.save_local(path=path, file_name=file_name)
                st.markdown(
                    '<p style="color:Green;">Feedback saved & sent to the Data Science team.</p>',
                    unsafe_allow_html=True,
                )

    def _collect_single_edge_case(self) -> Tuple[Union[str, int, None], str]:
        """
        Collect correct prediction for the single edge case flag.
        """
        st.write(
            self.__feedback_type_description(
                "you are signalling that the combination of all features is a critical edge case that we must test for."
            )
        )
        corrected_prediction = st.selectbox(
            "The model prediction for this edge case should be:",
            tuple(self.trubrics_model.data.testing_data[self.trubrics_model.data.target].unique()),
        )
        description = "A single edge case."
        return corrected_prediction, description

    def _collect_important_features_feedback(self) -> Tuple[str, int, str]:
        st.write(
            self.__feedback_type_description(
                "you are signalling that a given feature must be in the top N most important features."
            )
        )
        features = self.trubrics_model.data.features
        selected_feature = st.selectbox("Choose model feature:", (features))
        top_n_feature = st.slider(
            "The selected feature should be in the top ... features:", min_value=1, max_value=len(features)
        )
        description = "Most important features."
        return selected_feature, top_n_feature, description

    @staticmethod
    def __feedback_type_description(error_description: str) -> str:
        return f"Feedback type description: {error_description}"



import pandas as pd


def age_young(df: pd.DataFrame) -> pd.DataFrame:
    return df.loc[df["Age"] < 4, :]


def sex_female(df: pd.DataFrame) -> pd.DataFrame:
    return df.loc[df["Sex"] == "female", :]


def sex_male(df: pd.DataFrame) -> pd.DataFrame:
    return df.loc[df["Sex"] == "male", :]


slicing_functions = {"children": age_young, "female": sex_female, "male": sex_male}



import joblib
import pandas as pd
import streamlit as st
from sklearn.metrics import SCORERS

from trubrics.context import DataContext
from trubrics.validations import ModelValidator, Trubric
from trubrics.validations.run import TrubricRun, run_trubric


def expand_docstring(func):
    expander = st.expander("See docstring")
    stripped_docstring = "\n".join(" ".join(line.split()) for line in func.__doc__.split("\n"))
    expander.text(stripped_docstring)


st.set_page_config(layout="wide")

st.sidebar.title("Initialise a ModelValidator")
test_data = st.sidebar.file_uploader("Upload a test dataset")
train_data = st.sidebar.file_uploader("Upload a train dataset")
if test_data and train_data:
    test_df = pd.read_csv(test_data)
    train_df = pd.read_csv(train_data)
    st.sidebar.write("Preview of uploaded data:", test_df.head())
    target = st.sidebar.selectbox("Select the target variable from the uploaded test dataset:", test_df.columns)
    data_context = DataContext(
        testing_data=test_df,
        training_data=train_df,
        target=target,
    )

model_ = st.sidebar.file_uploader("Upload a model")
if model_ is not None:
    model = joblib.load(model_)

if model_ and test_data and train_data:
    model_validator = ModelValidator(data=data_context, model=model)  # type: ignore

    tab1, tab2 = st.tabs(["Select Model Validations", "Save Trubric"])
    with tab1:
        validations = []

        col1, col2, col3 = st.columns([2, 1.4, 1])
        with col1:
            perf_thresh = st.checkbox("validate_performance_against_threshold")
        with col2:
            metric = st.selectbox("1. Metric", SCORERS)
        with col3:
            threshold = st.number_input("1. Threshold", min_value=-1.0, max_value=1.0, value=0.2, step=0.1)
        expand_docstring(model_validator.validate_performance_against_threshold)
        if perf_thresh:
            perf_vs_thresh = model_validator.validate_performance_against_threshold(
                metric=metric,
                threshold=threshold,
            )
            validations.append(perf_vs_thresh)
            st.write(perf_vs_thresh.dict())

        col1, col2, col3 = st.columns([2, 1.4, 1])
        with col1:
            perf_dum = st.checkbox("validate_test_performance_against_dummy")
        with col2:
            metric = st.selectbox("2. Metric", SCORERS)
        with col3:
            strategy = st.selectbox("2. Strategy", ["most_frequent", "prior", "stratified", "uniform", "constant"])
            if strategy == "constant":
                cst_value = st.text_input("Enter constant value")
                if cst_value.isnumeric():
                    cst_value = int(cst_value)
                dummy_kwarg = {"metric": metric, "strategy": strategy, "dummy_kwargs": {"constant": cst_value}}
            else:
                dummy_kwarg = {"metric": metric, "strategy": strategy}
        expand_docstring(model_validator.validate_test_performance_against_dummy)
        if perf_dum:
            perf_vs_dummy = model_validator.validate_test_performance_against_dummy(**dummy_kwarg)
            validations.append(perf_vs_dummy)
            st.write(perf_vs_dummy.dict())

        col1, col2, col3 = st.columns([2, 1.4, 1])
        with col1:
            perf_train = st.checkbox("validate_performance_between_train_and_test")
        with col2:
            metric = st.selectbox("3. Metric", SCORERS)
        with col3:
            threshold = st.number_input("3. Threshold", min_value=-1.0, max_value=1.0, value=0.2, step=0.1)
        expand_docstring(model_validator.validate_performance_between_train_and_test)
        if perf_train:
            perf_vs_train = model_validator.validate_performance_between_train_and_test(
                metric=metric, threshold=threshold
            )
            validations.append(perf_vs_train)
            st.write(perf_vs_train.dict())

        col1, col2, col3 = st.columns([2, 1.4, 1])
        with col1:
            inf_time = st.checkbox("validate_inference_time")
        with col2:
            threshold = st.number_input("4. Threshold (s)", min_value=0.0, max_value=100.0, value=0.1, step=0.1)
        with col3:
            n_executions = st.number_input("4. n_executions", min_value=1, max_value=1000, value=100, step=50)
        expand_docstring(model_validator.validate_inference_time)
        if inf_time:
            inference_time = model_validator.validate_inference_time(threshold=threshold, n_executions=n_executions)
            validations.append(inference_time)
            st.write(inference_time.dict())

        with tab2:
            tc = Trubric(data_context_name="data", data_context_version=0.1, validations=validations)

            path = st.text_input("Filepath to save trubric to:")
            if st.button("Save trubric locally"):
                tc.save_local(path)
                msg = f'<p style="color:Green;">Trubric saved to {path}/my_trubric.json!</p>'
                st.markdown(msg, unsafe_allow_html=True)

            if st.button("Run saved trubric"):
                trubric = Trubric.parse_file(f"{path}/my_trubric.json")

                run_context = TrubricRun(
                    data_context=data_context,  # type: ignore
                    model=model,  # type: ignore
                    trubric=trubric,
                )

                all_validation_results = run_trubric(run_context)

                for validation_result in all_validation_results:
                    message_start = f"{validation_result.validation_type} - {validation_result.severity.upper()}"
                    completed_dots = (100 - len(message_start)) * "."
                    if validation_result.outcome == "pass":
                        msg = f'<p style="color:Green;">{message_start + completed_dots + "PASSED"}</p>'
                    else:
                        msg = f'<p style="color:Red;">{message_start + completed_dots + "FAILED"}</p>'
                    st.markdown(msg, unsafe_allow_html=True)



from pathlib import Path
from typing import Any, Dict, List, Optional

from loguru import logger
from pydantic import BaseModel, validator

from trubrics.utils.trubrics_manager_connector import make_request


def _validation_context_example():
    return {
        "example": {
            "validation_type": "validate_performance_against_threshold",
            "validation_kwargs": {"args": [], "kwargs": {"threshold": 0.8}},
            "outcome": "fail",
            "severity": "error",
            "result": {"performance": "0.79"},
        }
    }


class Validation(BaseModel):
    """
    Dataclass for a single validation point. Must be serialisable to .json, as is fed into Trubric dataclass.

    Note:
        A Validation object constrains the output of validations, with the @validation_output decorator.

    Attributes:
        validation_type: method name of the validation.
        validation_kwargs: all args and kwargs that the validation had run with.
        explanation: docstring explanation of the validation.
        outcome: pass or fail output of the validation.
        severity: severity of the validation, can be one of ["error", "warning", "experiment"], is "error" by default
        result: a dictionary of contextual elements calculated during the validation run
    """

    validation_type: str
    validation_kwargs: Dict[str, Optional[Any]]
    explanation: str
    outcome: str
    severity: str = "error"
    result: Optional[Dict[str, Optional[Any]]]

    class Config:
        extra = "forbid"
        validate_assignment = True
        schema_extra = _validation_context_example()

    @validator("severity")
    def severity_must_be(cls, v: str):
        severity_values = ["error", "warning", "experiment"]
        if v not in severity_values:
            raise KeyError(f"Severity must be set to: {severity_values}.")
        return v

    @validator("outcome")
    def outcome_must_be(cls, v: str):
        outcome_values = ["pass", "fail"]
        if v not in outcome_values:
            raise KeyError(f"Outcome must be set to: {outcome_values}.")
        return v


class Trubric(BaseModel):
    """
    Dataclass for a trubric, or set of validation points. Must be serialisable to .json.

    Attributes:
        name: Trubric name.
        model_name: model name
        model_version: model version
        data_context_name: data context name (from DataContext)
        data_context_version: data context version (from DataContext)
        metadata: free textual metadata field
        validations: list of validations (defined by Validation)
    """

    trubric_name: str = "my_trubric"
    model_name: str = "my_model"
    model_version: float = 0.1
    data_context_name: str
    data_context_version: float
    metadata: Optional[Dict[str, str]] = None
    validations: List[Validation]

    class Config:
        schema_extra = {
            "example": {
                "name": "my_first_trubric",
                "model_name": "my_model",
                "model_version": 0.1,
                "data_context_name": "my_dataset",
                "data_context_version": 0.1,
                "metadata": {},
                "validations": [_validation_context_example()],
            }
        }

    def save_local(self, path: str, file_name: Optional[str] = None):
        if path is None:
            raise TypeError("Specify the local path where you would like to save your Trubric json.")
        if file_name is None:
            file_name = f"{self.trubric_name}.json"
        with open(Path(path) / file_name, "w") as file:
            file.write(self.json(indent=4))
            logger.info(f"Trubric saved to {Path(path) / file_name}.")

    def save_ui(self, url: str, user_id: str):

        if user_id is None:
            raise TypeError("You must specify a 'user_id' to push to the trubrics manager.")
        else:
            make_request(
                f"{url}/api/trubrics/{user_id}",
                headers={"Content-Type": "application/json"},
                data=self.json().encode("utf-8"),
            )
            logger.info("Trubric saved to the trubrics manager.")



from trubrics.context import DataContext
from trubrics.validations import ModelValidator
from trubrics.validations.validation_output import (
    validation_output,
    validation_output_type,
)


class CustomValidator(ModelValidator):
    def __init__(self, data: DataContext, model, custom_scorers=None, slicing_functions=None):
        super().__init__(data, model, custom_scorers, slicing_functions)

    @validation_output
    def validate_performance_for_different_fares(self, fare_cutoff: int, severity=None):
        """Validate the models performance for expensive vs cheap tickets.

        Args:
            fare_cutoff: value to split the fare column on

        Returns:
            True for success, false otherwise. With a results dictionary giving all the number of errors by fare split.
        """
        return self._validate_performance_for_different_fares(fare_cutoff)

    def _validate_performance_for_different_fares(self, fare_cutoff: int = 50) -> validation_output_type:
        """
        Write your custom validation function here.

        Notes
        -----
            This method is separated from validate_performance_for_different_fares
            to apply @validation_output and for unit testing.

            The @validation_output decorator allows you to generate a Validation object,
            and must be used to be able to save your validation as part of a Trubric.
            This decorator requires you to return values with the same type as validation_output_type.
        """

        errors_df = self.tm.testing_data_errors
        number_of_errors_by_fare_ratio = (
            errors_df.loc[lambda x: x["Fare"] <= fare_cutoff].shape[0]
            / errors_df.loc[lambda x: x["Fare"] > fare_cutoff].shape[0]
        )
        return (
            number_of_errors_by_fare_ratio > 0.5 and number_of_errors_by_fare_ratio < 1.5,
            {"number_of_errors_by_fare_ratio": round(number_of_errors_by_fare_ratio, 3)},
        )



from urllib.request import Request, urlopen


def make_request(url, headers=None, data=None):
    request = Request(url, headers=headers or {}, data=data)
    with urlopen(request, timeout=20) as response:
        return response.read()
